"""
Defines data structures and routines for dealing with a pokemon's move
"""

import random
import enum
import os
from . import constants
from . import poketypes
from . import utils


def critical(movecrit: int, pokestages:int) -> float:
	"""
	Given a move's inherent critical hit ratio and a pokemon's critical stage,
	returns a multiplier to apply to damage from critical hits
	"""
	effectivestage = (movecrit + pokestages)

	if effectivestage > 3:
		return 1.5

	ratio = 1.0 / (2 ** (4-effectivestage))
	critChance = random.uniform(0,1)

	return 1.5 if ratio > critChance else 1

class MoveType(enum.IntEnum):
	"""
	The type of a move
	"""
	Status   = 0
	Special  = 1
	Physical = 2

	def __str__(self) -> str:
		"""
		Returns the name of a move type
		"""
		return self._name_

# Move Type constants for export
STATUS = MoveType.Status
SPECIAL = MoveType.Special
PHYSICAL = MoveType.Physical

class Move():
	"""
	A pokemon's move
	"""

	def __init__(self, name: str):
		self.name = name
		self.priority = 0
		self.type1 = poketypes.TYPELESS
		self.type2 = poketypes.TYPELESS
		self.contact = False
		self.accuracy = 0
		self.pp = 0
		self.maxPP = 0
		self.moveType = MoveType.Status
		self.crit = 0
		self.power = 0
		self.target = False
		self.stages = []
		self.affectedStats = []

	def readData(self):
		"""
		Reads in a move's data. Expects it in `../data/moves/<name>`
		"""

		with open(os.path.join(utils.dataDir, "moves", self.name)) as datafile:
			lines = datafile.read().split('\n')
			_ = lines.pop() #eliminates POSIX-compliant empty line at file end
			self.type1 = poketypes.Type(int(lines.pop(0)))
			self.type2 = poketypes.Type(int(lines.pop(0)))
			self.contact = bool(int(lines.pop(0)))
			self.accuracy = int(lines.pop(0))
			self.pp = int(lines.pop(0))
			self.maxPP = self.pp
			self.moveType = MoveType(int(lines.pop(0)))

		if self.moveType != STATUS:
			self.damagingMove(lines)
		else:
			self.statusMove(lines)

	def damagingMove(self, lines: list[str]):
		"""
		Builds a damaging move from the remaining lines in a file after common
		attributes have already been parsed out
		"""
		self.crit = int(lines.pop())
		self.power = int(lines.pop())

	def statusMove(self, lines: list[str]):
		"""
		Builds a status move from the remaining lines in a file after common
		attributes have already been parsed out
		"""
		self.target = bool(int(lines.pop()))
		self.stageChanges = [int(stage) for stage in lines.pop().split(' ')]
		self.affectedStats = [constants.Stats(int(stat)) for stat in lines.pop().split(' ')]

	def __repr__(self) -> str:
		"""
		Forms an in-depth representation of a Move
		"""
		ret = [str(self)]
		ret.append(f"Type: {self.type1}{'/'+str(self.type2) if self.type2 != poketypes.TYPELESS else ''}")
		ret.append(str(self.moveType))
		ret.append(f"Accuracy: {self.accuracy}")
		if self.moveType != STATUS:
			ret.append(f"Power: {self.power}")
			ret.append(f"Does {'not ' if not self.contact else ''}make contact")

		return "\n".join(ret)

	def __str__(self) -> str:
		"""
		Returns the human-readable name of a Move
		"""
		return self.name
